import numpy as np
from geopy.distance import distance
from scipy import interpolate
import matplotlib.pyplot as plt


def points_track_dist(kla, klo, tla, tlo):
    K, N = len(klo), len(tlo)
    d = np.zeros((K, N))
    for i in range(K):
        for j in range(N):
            d[i, j] = distance((kla[i], klo[i]), (tla[j], tlo[j])).m
    return d
    

def write_file(tlo, tla, tn, last_point):
    f = open('track_with_chn.out', 'w')
    j = 0
    for j in range(last_point+1):
        f.write('{:.10f}   {:.10f}   {:d}\n'.format(tlo[j], tla[j], int(tn[j])))
    f.close()
    return None


def intrack_dist(tlo, tla):
    K = len(tlo)
    d = np.zeros(K)
    for i in range(1, K):
        d[i] = distance((tla[i-1], tlo[i-1]), (tla[i], tlo[i])).m
    return d
    

def get_channel_loc(tlo, tla, tn, d_inv):
    l_track = intrack_dist(tlo, tla) 
    l_track_cum = np.cumsum(l_track)
    idx_kp = np.where(tn>=0)[0]

    chn = 0
    f = open('interp_channel.txt', 'w')
    f_sr = open('sag_ratio.txt', 'w')
    for i in range(1, len(idx_kp)):

        # calculate actual interval between known-channel points 
        istart, iend = idx_kp[i-1], idx_kp[i]+1
        n_chn_kp = tn[iend-1]-tn[istart]
        d_interp = (l_track_cum[iend-1]-l_track_cum[istart]) / n_chn_kp
        f_sr.write('{:5d} {:5d} {:10.3f} {:10.3f} {:10.3f}\n'.format(
                        int(tn[istart]), int(tn[iend-1]),
                        d_interp, d_inv, 1-d_interp/d_inv))
        
        l_res = 0 # remaining fiber length before counting the next segment

        for j in range(istart, iend-1):
            l_start = l_track[j+1] + l_res

            # if tp segment length is large for more than one interval, get the channel loc
            if l_start >= d_interp:
                n_chn_tp = int(l_start/d_interp) # floor int, num of channel available 
                x_new = (np.arange(n_chn_tp)+1)*d_interp-l_res # channel distance from segment start
  
                # interpolate the channel loc
                flo = interpolate.interp1d([0, l_track[j+1]], [tlo[j], tlo[j+1]])
                fla = interpolate.interp1d([0, l_track[j+1]], [tla[j], tla[j+1]])
                tlo_new, tla_new = flo(x_new), fla(x_new)

                # remaining length to add to next segment
                l_res = l_start-n_chn_tp*d_interp
               
                # write interpolated channel loc
                for k in range(tlo_new.size):
                    chn += 1
                    f.write('{:.10f} {:.10f} {:10d}\n'.format(tlo_new[k], tla_new[k], chn))

                # handle floor int problem when l_start/d_interp is near an interger (e.g. 3.999)
                if (d_interp-l_res)/d_interp < 1.0e-6:
                    chn += 1
                    f.write('{:.10f} {:.10f} {:10d}\n'.format(tlo[j+1], tla[j+1], int(tn[j+1]), chn))
                    l_res = 0
            # if tp segment length is not enough for one interval, simply add the length to next segment        
            elif l_start < d_interp:
                l_res = l_track[j+1] 

    f.close()
    f_sr.close()
    print('Interpolated channels written to interp_channel.txt')
    print('Sag ratio between consecutive known-channel points written to sag_ratio.txt')

    return None


if __name__ == '__main__':

    # load fiber spatial track and points with known chn numbers
    klo, kla, kn = np.loadtxt('known_points.txt', unpack=True)
    tlo, tla = np.loadtxt('fiber_track.txt', unpack=True)
    tn = np.zeros(len(tlo))-1.0
    d_inv = 4 #channel interval in m

    
    # insert the known points into the fiber track data
    # points far from the track (> chn interval) are excluded
    K, N = len(klo), len(tlo)
    d = points_track_dist(kla, klo, tla, tlo)
    idx = np.argmin(d, axis=1)
    for i in range(K):
        if d[i, idx[i]] < d_inv: 
            tn[idx[i]] = kn[i]
            last_point = idx[i]
    write_file(tlo, tla, tn, last_point)

    
    # interpolation with regular spacing along the fiber track
    tlo, tla, tn = np.loadtxt('track_with_chn.out', unpack=True)
    get_channel_loc(tlo, tla, tn, d_inv)


    # map for verification
    chlo, chla, chn = np.loadtxt('interp_channel.txt', unpack=True)
    plt.plot(tlo, tla, 'k')
    plt.plot(chlo, chla, 'r+')
    idx_kp = np.where(tn>=0)[0]
    for i in idx_kp[1:]:
        ichn = int(tn[i])
        plt.plot(tlo[i], tla[i], 'ko')
        plt.text(tlo[i], tla[i], 'track '+str(ichn), verticalalignment='top')
        plt.plot(chlo[ichn-1], chla[ichn-1], 'r+') # minus 1 due to python start with 0 but channel start with 1
        plt.text(chlo[ichn-1], chla[ichn-1], 'interp '+str(int(chn[ichn-1])), verticalalignment='bottom')
    #plt.savefig('map.png')
    plt.show()
    print('Illustration map written map.png')
